/*
 * Copyright (c) 2003 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: worker.c,v 1.2 2003/06/06 19:38:05 fuyu Exp $
 */

#include "config.h"

#include "ochusha_private.h"
#include "ochusha.h"

#include "worker.h"

#include <glib.h>
#include <ghttp.h>

#include <pthread.h>
#include <stdio.h>
#include <stdlib.h>

static pthread_mutex_t job_list_lock;
static volatile GSList *job_list;

static volatile GSList *worker_threads;
static int worker_threads_max;
static volatile int number_of_threads;
static volatile int number_of_idle_threads;
static pthread_cond_t idle_threads;

static pthread_attr_t worker_thread_attribute;

static WorkerThread *worker_thread_new(void);
static void *worker_thread_main(void *args);


#define JOB_LIST_LOCK					\
  if (pthread_mutex_lock(&job_list_lock) != 0)		\
    {							\
      fprintf(stderr, "Couldn't lock a mutex.\n");	\
      abort();						\
    }

#define JOB_LIST_UNLOCK					\
  if (pthread_mutex_unlock(&job_list_lock) != 0)	\
    {							\
      fprintf(stderr, "Couldn't unlock a mutex.\n");	\
      abort();						\
    }


#define NEED_LARGER_STACK	0


void
initialize_worker(int initial_num_threads, int maximum_num_threads)
{
  int i;
#if NEED_LARGER_STACK
  size_t stack_size;
#endif

  job_list = NULL;
  worker_threads = NULL;
  if (maximum_num_threads > 0)
    worker_threads_max = maximum_num_threads;
  else
    worker_threads_max = 0x7fffffff;	/* ޤǡ */
  number_of_threads = initial_num_threads;
  number_of_idle_threads = 0;

  if (maximum_num_threads < initial_num_threads)
    {
      fprintf(stderr,
	      "initialize_worker: Invalid args(maximum(%d) < initial(%d).\n",
	      maximum_num_threads, initial_num_threads);
      abort();
    }

  if (pthread_mutex_init(&job_list_lock, NULL) != 0)
    {
      fprintf(stderr, "Couldn't init a mutex.\n");
      abort();
    }

  if (pthread_cond_init(&idle_threads, NULL) != 0)
    {
      fprintf(stderr, "Couldn't init a condition variable.\n");
      abort();
    }

  if (pthread_attr_init(&worker_thread_attribute) != 0)
    {
      fprintf(stderr, "Couldn't init a worker threads' attribute.\n");
      abort();
    }

#if NEED_LARGER_STACK
  /* å­ʤΤȻפä㤦ХǤ뤳ȤȽΤǼΤ */
  if (pthread_attr_getstacksize(&worker_thread_attribute, &stack_size) != 0)
    {
      fprintf(stderr, "Couldn't get default stack_size.\n");
    }
  else
    fprintf(stderr, "default stack_size = 0x%x\n", stack_size);

  if (pthread_attr_setstacksize(&worker_thread_attribute, stack_size * 2) != 0)
    {
      fprintf(stderr, "Couldn't set worker threads' stack_size as 0x%x.\n",
	      stack_size * 2);
      abort();
    }
#endif

  for (i = 0; i < initial_num_threads; i++)
    worker_thread_new();
}


static void
set_die_now_command_for_all_threads(gpointer data, gpointer unused)
{
  WorkerThread *thread = (WorkerThread *)data;
  thread->command = DIE_NOW;
}


void
terminate_workers(void)
{
  GSList *list_entry;
  g_slist_foreach((GSList *)worker_threads,
		  set_die_now_command_for_all_threads, NULL);

  while ((list_entry = (GSList *)worker_threads) != NULL)
    {
      WorkerThread *thread;
      worker_threads = g_slist_remove_link((GSList *)worker_threads,
					   list_entry);
      thread = (WorkerThread *)list_entry->data;
      g_slist_free_1(list_entry);
      JOB_LIST_LOCK
	{
	  if (pthread_cond_broadcast(&idle_threads) != 0)
	    {
	      fprintf(stderr, "Couldn't broadcast a condition variable.\n");
	      abort();
	    }
	}
      JOB_LIST_UNLOCK;

      if (pthread_join(thread->id, NULL) != 0)
	{
	  fprintf(stderr, "Couldn't join a thread\n");
	  abort();
	}

      G_FREE(thread);
    }
}


static WorkerThread *
worker_thread_new(void)
{
  WorkerThread *thread = G_NEW0(WorkerThread, 1);
  thread->command = GO_AHEAD;

  if (pthread_create(&thread->id, &worker_thread_attribute,
		     worker_thread_main, thread) != 0)
    {
      fprintf(stderr, "Couldn't create a worker thread.\n");
      abort();
    }

  worker_threads = g_slist_append((GSList *)worker_threads, thread);
  number_of_threads++;

  return thread;
}


void
commit_job(WorkerJob *job)
{
  JOB_LIST_LOCK
    {

#if DEBUG_THREAD_MOST
      fprintf(stderr, "commit_job: job committed.\n");
      fflush(stderr);
#endif

      job_list = g_slist_append((GSList *)job_list, job);

      if (number_of_idle_threads > 0)
	{
	  number_of_idle_threads--;

#if DEBUG_THREAD_MOST
	  fprintf(stderr, "signaling to awake idle thread.\n");
	  fflush(stderr);
#endif

	  if (pthread_cond_signal(&idle_threads) != 0)
	    {
	      fprintf(stderr, "Couldn't signal a condition variable.\n");
	      abort();
	    }

	  JOB_LIST_UNLOCK;
#if DEBUG_THREAD_MOST
	  fprintf(stderr, "Job must be started!\n");
	  fflush(stderr);
#endif
	  return;
	}

#if DEBUG_THREAD_MOST
      fprintf(stderr, "There's no idle thread.\n");
      fflush(stderr);
#endif

      if (number_of_threads < worker_threads_max)
	worker_thread_new();
    }
  JOB_LIST_UNLOCK;
}


static void *
worker_thread_main(void *args)
{
  WorkerThread *self = (WorkerThread *)args;
  JOB_LIST_LOCK
    {
#if DEBUG_THREAD_MOST
      fprintf(stderr, "WorkerThread is created.\n");
      fflush(stderr);
#endif

      while (1)
	{
	  GSList *list_entry;
	  WorkerJob *job;
	  if (self->command == DIE_NOW)
	    {
	      JOB_LIST_UNLOCK;
#if DEBUG_THREAD_MOST
	      fprintf(stderr, "WorkerThread is killed.\n");
	      fflush(stderr);
#endif
	      pthread_exit(NULL);
	    }

	  list_entry = (GSList *)job_list;
	  if (list_entry == NULL)
	    {
#if DEBUG_THREAD_MOST
	      fprintf(stderr, "There's no job to be done.\n");
	      fflush(stderr);
#endif
	      number_of_idle_threads++;
	      pthread_cond_wait(&idle_threads, &job_list_lock);
	      continue;
	    }

	  job_list = g_slist_remove_link((GSList *)job_list, list_entry);
	  job = (WorkerJob *)list_entry->data;
	  g_slist_free_1(list_entry);

	  JOB_LIST_UNLOCK;

	  (*job->job)(self, job->args);
	  G_FREE(job);
	  JOB_LIST_LOCK;
	}
    }
}
